package jalview.gui;

import jalview.datamodel.*;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import java.awt.image.*;

public class AnnotationPanel extends JPanel implements MouseListener, MouseMotionListener, ActionListener, AdjustmentListener
{
  AlignViewport av;
  AlignmentPanel ap;
  int activeRow =-1;

  ArrayList activeRes;
  static String HELIX ="Helix";
  static String SHEET ="Sheet";
  static String LABEL ="Label";
  static String REMOVE="Remove Annotation";
  static String COLOUR="Colour";
  static Color HELIX_COLOUR = Color.red.darker();
  static Color SHEET_COLOUR = Color.green.darker().darker();


  BufferedImage image;
  Graphics2D gg;
  FontMetrics fm;
  int imgWidth=0;

  boolean fastPaint = false;

  public static int GRAPH_HEIGHT = 40;



  public AnnotationPanel(AlignmentPanel ap)
  {
    this.ap = ap;
    av = ap.av;
    this.setLayout(null);
    addMouseListener(this);
    addMouseMotionListener(this);
    adjustPanelHeight();


    ap.annotationScroller.getVerticalScrollBar().addAdjustmentListener( this );
  }

  public void adjustmentValueChanged(AdjustmentEvent evt)
  {
    ap.alabels.setScrollOffset( -evt.getValue() );
  }

  public void adjustPanelHeight()
  {
    // setHeight of panels
    AlignmentAnnotation[] aa = av.alignment.getAlignmentAnnotation();
    int height = 0;
    if(aa!=null)
    for (int i = 0; i < aa.length; i++)
    {
      if(!aa[i].visible)
        continue;

      aa[i].height = 0;

      if(aa[i].hasText)
        aa[i].height += av.charHeight;
      if (aa[i].hasIcons)
        aa[i].height += 16;

      if (aa[i].isGraph)
        aa[i].height += GRAPH_HEIGHT;

      if(aa[i].height==0)
        aa[i].height = 20;
      height += aa[i].height;
    }
  else height=20;
    this.setPreferredSize(new Dimension(1, height));

  }

  public void addEditableColumn(int i)
  {
    if(activeRow==-1)
    {
      AlignmentAnnotation [] aa = av.alignment.getAlignmentAnnotation();
      for(int j=0; j<aa.length; j++)
        if(aa[j].editable)
        {
          activeRow = j;
          break;
        }
    }

    if(activeRes==null)
    {
      activeRes = new ArrayList();
      activeRes.add(String.valueOf(i));
      return;
    }

    activeRes.add(String.valueOf(i));

  }


  public void actionPerformed(ActionEvent evt)
  {

    AlignmentAnnotation [] aa = av.alignment.getAlignmentAnnotation();
    Annotation [] anot = aa[activeRow].annotations;

    if(evt.getActionCommand().equals(REMOVE))
    {
      for(int i=0; i<activeRes.size(); i++)
      {
        anot[Integer.parseInt(activeRes.get(i).toString())] = null;
        anot[Integer.parseInt(activeRes.get(i).toString())] = null;
      }
    }
    else if(evt.getActionCommand().equals(LABEL))
    {
      String label = JOptionPane.showInputDialog(this, "Enter Label ", "Enter label", JOptionPane.QUESTION_MESSAGE );
      if(label==null)
        label = "";

      if(label.length()>0 && !aa[activeRow].hasText)
       aa[activeRow].hasText = true;

      for(int i=0; i<activeRes.size(); i++)
      {
        int index = Integer.parseInt(activeRes.get(i).toString());
        if(anot[index]==null)
          anot[index] = new Annotation(label, "", ' ',0);
        anot[index].displayCharacter = label;
      }
    }
    else if(evt.getActionCommand().equals(COLOUR))
    {
     Color col = JColorChooser.showDialog(this, "Choose foreground colour", Color.black);
     for (int i = 0; i < activeRes.size(); i++)
     {
       int index = Integer.parseInt(activeRes.get(i).toString());
       if (anot[index] == null)
         anot[index] = new Annotation("", "", ' ', 0);
       anot[index].colour = col;
     }
    }
    else // HELIX OR SHEET
    {
    char type = 0;
    String symbol = "\u03B1";
    if(evt.getActionCommand().equals(HELIX))
       type = 'H';
    else if(evt.getActionCommand().equals(SHEET))
    {
      type = 'E';
      symbol = "\u03B2";
    }

    if(!aa[activeRow].hasIcons)
      aa[activeRow].hasIcons = true;


    String label = JOptionPane.showInputDialog("Enter a label for the structure?", symbol );
    if(label==null)
       label="";

    if(label.length()>0 && !aa[activeRow].hasText)
       aa[activeRow].hasText = true;

    for(int i=0; i<activeRes.size(); i++)
    {
      int index = Integer.parseInt(activeRes.get(i).toString());
      if (anot[index] == null)
      {
        anot[index] = new Annotation(label, "", type, 0);
      }

        anot[ index ].secondaryStructure = type;
        anot[ index ].displayCharacter = label;
    }
    }

    adjustPanelHeight();
    activeRes = null;
    repaint();
    return;


  }

  public void mousePressed(MouseEvent evt)
  {
    if (SwingUtilities.isRightMouseButton(evt))
    {
      if(activeRes==null)
        return;

      JPopupMenu pop = new JPopupMenu("Structure type");
      JMenuItem item = new JMenuItem(HELIX);
      item.addActionListener(this);
      pop.add(item);
      item = new JMenuItem(SHEET);
      item.addActionListener(this);
      pop.add(item);
      item = new JMenuItem(LABEL);
      item.addActionListener(this);
      pop.add(item);
      item = new JMenuItem(COLOUR);
      item.addActionListener(this);
      pop.add(item);
      item = new JMenuItem(REMOVE);
      item.addActionListener(this);
      pop.add(item);
      pop.show(this, evt.getX(), evt.getY());
      return;
    }


    AlignmentAnnotation [] aa = av.alignment.getAlignmentAnnotation();
    if(aa==null)
      return;

    int height = 0;
    activeRow = -1;
    for(int i=0; i<aa.length; i++)
    {
      height+= aa[i].height;

      if(evt.getY()<height)
      {
        if(!aa[i].editable)
        {
          activeRes = null;
          continue;
        }

        activeRow = i;
        break;
      }
    }


    int res = evt.getX() / av.getCharWidth() + av.getStartRes();

    if(evt.isControlDown() || evt.isAltDown())
      addEditableColumn(res);

    else if(evt.isShiftDown())
    {

      if(activeRes==null)
        activeRes=new ArrayList();
      else
      {
          int start =  Integer.parseInt( activeRes.get( activeRes.size()-1 ).toString() );
          int end = res;
          if(end<start)
          {
            int temp = end;
            end = start;
            start = temp;
          }
          for(int n=start; n<=end; n++)
            addEditableColumn(n);

      }
    }
    else
    {
        activeRes = new ArrayList();
        activeRes.add( String.valueOf(res) );
    }

    repaint();

  }
  public void mouseReleased(MouseEvent evt)
  {  }
  public void mouseEntered(MouseEvent evt)
  {  }
  public void mouseExited(MouseEvent evt)
  {  }
  public void mouseDragged(MouseEvent evt)
  {  }
  public void mouseMoved(MouseEvent evt)
  {
    ToolTipManager.sharedInstance().registerComponent(this);
    AlignmentAnnotation [] aa = av.alignment.getAlignmentAnnotation();
    if(aa==null)
      return;

    int row = -1;
    int height=0;
    for(int i=0; i<aa.length; i++)
    {

      if( aa[i].visible )
        height += aa[i].height;

      if(evt.getY()<height)
      {
        row = i;
        break;
      }
    }

    int res = evt.getX() / av.getCharWidth() + av.getStartRes();
    if(row>-1 && res<aa[row].annotations.length && aa[row].annotations[res]!=null)
      this.setToolTipText(aa[row].annotations[res].description);

  }
  public void mouseClicked(MouseEvent evt) {}


  public void paintComponent(Graphics g)
  {
    g.setColor(Color.white);
    g.fillRect(0,0,getWidth(), getHeight());

    if(fastPaint)
    {
      g.drawImage(image, 0, 0, this);
      fastPaint = false;
      return;
    }

    imgWidth = (av.endRes-av.startRes+1) *av.charWidth;

    image = new BufferedImage(imgWidth,
                                ap.annotationPanel.getHeight(),
                                BufferedImage.TYPE_INT_RGB);
      gg = (Graphics2D) image.getGraphics();
      gg.setColor(Color.white);
      gg.fillRect(0, 0, imgWidth, getHeight());
      gg.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                          RenderingHints.VALUE_ANTIALIAS_ON);
      fm = gg.getFontMetrics();
      gg.setFont(av.getFont());


    drawComponent( gg, av.startRes, av.endRes+1);
    g.drawImage( image, 0, 0, this);
  }

  public void fastPaint(int horizontal)
{
  if(      horizontal ==0
     || av.alignment.getAlignmentAnnotation()==null
     || av.alignment.getAlignmentAnnotation().length<1
    )
  {
    repaint();
    return;
  }

  gg.copyArea( 0,0, imgWidth, getHeight(), -horizontal*av.charWidth, 0 );
  int sr=av.startRes, er=av.endRes+1, transX=0;

  if(horizontal>0) // scrollbar pulled right, image to the left
  {
    transX =  (er-sr-horizontal)*av.charWidth;
    sr = er - horizontal ;
  }
  else if(horizontal<0)
  {
    er = sr-horizontal;
  }


  gg.translate(transX, 0);

  drawComponent(gg, sr, er);

  gg.translate( -transX, 0 );

  fastPaint = true;
  repaint();
}


  public void drawComponent(Graphics2D g, int startRes, int endRes)
  {
    g.setColor(Color.white);
    g.fillRect(0,0,(endRes-startRes) *av.charWidth, getHeight());
    if(av.alignment.getAlignmentAnnotation()==null || av.alignment.getAlignmentAnnotation().length<1)
    {
      g.setColor(Color.white);
      g.fillRect(0,0,getWidth(), getHeight());
      g.setColor(Color.black);
      g.drawString("Alignment has no annotations",20,15);
      return;
    }

    AlignmentAnnotation [] aa = av.alignment.getAlignmentAnnotation();

    int j, x=0, y=0;
    char [] lastSS = new char[aa.length];
    int  [] lastSSX= new int[aa.length] ;
    int iconOffset = av.charHeight/2;
    boolean validRes = false;
    //\u03B2 \u03B1
    for(int i=0; i<aa.length; i++)
    {
      AlignmentAnnotation row = aa[i];
      if(!row.visible)
        continue;

      if(row.isGraph)
      {
        // this is so that we draw the characters below the graph
        y += row.height;
        if(row.hasText)
          y -= av.charHeight;
      }
      if(row.hasText)
        iconOffset = av.charHeight/2;
      else
        iconOffset = 0;

      for(j=startRes; j<endRes; j++)
      {
        validRes = row.annotations[j]==null?false:true;

       x = (j-startRes)*av.charWidth;


       if(activeRow==i)
       {

         g.setColor(Color.red);

         if(activeRes!=null)
           for (int n = 0; n < activeRes.size(); n++)
           {
             int v = Integer.parseInt(activeRes.get(n).toString()) ;
             if (v == j)
               g.fillRect( (j-startRes) * av.charWidth, y, av.charWidth, row.height);
           }
       }



       if(validRes && row.annotations[j].displayCharacter.length()>0)
       {
         int charOffset = (av.charWidth -
                            fm.charWidth(row.annotations[j].displayCharacter.
                                         charAt(0))) / 2;
         g.setColor( row.annotations[j].colour);
          if(j==0)
          {
            if (row.annotations[0].secondaryStructure == 'H'
                || row.annotations[0].secondaryStructure == 'E')
              g.drawString(row.annotations[j].displayCharacter, x,
                           y + iconOffset + 2);
          }
         else if( (row.annotations[j].secondaryStructure=='H'
               || row.annotations[j].secondaryStructure=='E') &&
               (row.annotations[j-1]==null ||
                row.annotations[j].secondaryStructure!=row.annotations[j-1].secondaryStructure))

        g.drawString(row.annotations[j].displayCharacter, x, y + iconOffset + 2);

        if(!row.hasIcons)
            g.drawString(row.annotations[j].displayCharacter, x + charOffset,
                         y + iconOffset + 2);
       }

       if(row.hasIcons)
       if(!validRes || row.annotations[j].secondaryStructure!=lastSS[i])
       {
         switch (lastSS[i])
         {
           case 'H':
             g.setColor(HELIX_COLOUR);
             g.fillRoundRect(lastSSX[i], y+4 + iconOffset, x-lastSSX[i], 7, 8, 8);
             break;
           case 'E':
             g.setColor(SHEET_COLOUR);
             g.fillRect(lastSSX[i], y + 4 + iconOffset, x-lastSSX[i]-4, 7);
             g.fillPolygon(new int[] {x - 4, x- 4, x }
                           , new int[]{y+ iconOffset, y + 14+ iconOffset, y + 8+ iconOffset}, 3);
             break;
           case 'C':
             break;
           default :
             g.setColor(Color.gray);
             g.fillRect(lastSSX[i], y+6+ iconOffset, x-lastSSX[i], 2);
             break;
         }

         if(validRes)
           lastSS[i] = row.annotations[j].secondaryStructure;
          else
            lastSS[i] = ' ';
         lastSSX[i] = x;
       }

       if (validRes && row.isGraph)
       {
         g.setColor(new Color(0,0,180));
         int height = (int)((row.annotations[j].value / row.graphMax)*GRAPH_HEIGHT);

         if(row.windowLength>1)
         {
           int total =0;
           for(int i2=j- (row.windowLength/2); i2<j+(row.windowLength/2); i2++)
           {
             if(i2<0 || i2>=av.alignment.getWidth())
               continue;

             total += row.annotations[i2].value;
           }

           total/=row.windowLength;
           height = (int)( (total / row.graphMax) *GRAPH_HEIGHT);

         }
         g.setColor(row.annotations[j].colour);
         g.fillRect(x, y-height, av.charWidth, height );
       }


      }

      x+=av.charWidth;

      if(row.hasIcons)
      switch (lastSS[i])
      {
        case 'H':
          g.setColor(HELIX_COLOUR);
          g.fillRoundRect(lastSSX[i], y+4+ iconOffset, x - lastSSX[i], 7, 8, 8);
          break;
        case 'E':
          g.setColor(SHEET_COLOUR);
          g.fillRect(lastSSX[i], y + 4+ iconOffset, x - lastSSX[i] - 4, 7);
          g.fillPolygon(new int[]
                        {x - 4, x - 4, x}
                        , new int[]
                        {y + iconOffset, y + 14+ iconOffset, y + 7+ iconOffset}
                        , 3);
          break;
        case 'C':
          break;
        default:
          g.setColor(Color.gray);
          g.fillRect(lastSSX[i], y+6+ iconOffset, x-lastSSX[i], 2);
          break;

      }

       if(row.isGraph && row.hasText)
         y+=av.charHeight;
       if(!row.isGraph)
          y+=aa[i].height;
    }
  }

  // used by overview window
  public void drawGraph(Graphics g, AlignmentAnnotation aa,int width, int y)
  {
    g.setColor(Color.white);
    g.fillRect(0,0,width, y);
    g.setColor(new Color(0,0,180));
    int x = 0;
    for(int j=0; j<aa.annotations.length; j++)
    {
      g.setColor(new Color(0, 0, 180));
      int height = (int) ( (aa.annotations[j].value / aa.graphMax) * GRAPH_HEIGHT );
      g.fillRect(x, y - height, av.charWidth, height);
      x+=av.charWidth;
    }
  }
}
